from iso7816.apdu import CommandCase, CommandField, ResponseField, CommandApdu, ResponseApdu
from iso7816.encodings import SecureMessaging, Chaining, CLA
from iso7816.commands import GET_RESPONSE, SELECT



"""gp.py
"""

# Standard library imports
from typing import Union

# Third party imports

# Local application imports
from common.ber import parse_to_dict
from globalplatform import SELECT as _SELECT, GET_DATA as _GET_DATA, FileOccurrence, ApplicationIdentifier, SecureMessaging, GetDataObject
from globalplatform.encodings import CPLC, CardPersonalizationLifeCycleData
from pcsc.scard import Protocol, send_apdu


def SELECT(hcard, protocol: Protocol, logical_channel: int, file_occurrence: FileOccurrence, application_identifier: ApplicationIdentifier):
    return send_apdu(hcard, protocol, _SELECT(logical_channel, file_occurrence, application_identifier))


def GET_DATA(hcard, protocol: Protocol, secure_messaging: SecureMessaging, logical_channel: int, tag: Union[str, GetDataObject]):
    return send_apdu(hcard, protocol, _GET_DATA(secure_messaging, logical_channel, tag))


def GET_CPLC(hcard, protocol: Protocol, secure_messaging: SecureMessaging, logical_channel: int):
    r = send_apdu(hcard, protocol, _GET_DATA(secure_messaging,
                  logical_channel, GetDataObject.CardProductionLifeCycle))

    cplc_data = CPLC(parse_to_dict(r.data()).get('9F7F', {}))

    return cplc_data



"""scard.py
"""

# Standard library imports
import logging
from enum import IntEnum

# Third party imports
from asyncore import read
from smartcard.scard import \
    SCARD_SCOPE_USER as _SCARD_SCOPE_USER,\
    SCARD_SCOPE_SYSTEM as _SCARD_SCOPE_SYSTEM,\
    SCARD_SHARE_SHARED as _SCARD_SHARE_SHARED,\
    SCARD_SHARE_EXCLUSIVE as _SCARD_SHARE_EXCLUSIVE,\
    SCARD_SHARE_DIRECT as _SCARD_SHARE_DIRECT,\
    SCARD_PROTOCOL_T0 as _SCARD_PROTOCOL_T0,\
    SCARD_PROTOCOL_T1 as _SCARD_PROTOCOL_T1,\
    SCARD_PCI_T0 as _SCARD_PCI_T0,\
    SCARD_PCI_T1 as _SCARD_PCI_T1,\
    SCARD_PROTOCOL_RAW as _SCARD_PROTOCOL_RAW,\
    SCARD_LEAVE_CARD as _SCARD_LEAVE_CARD,\
    SCARD_RESET_CARD as _SCARD_RESET_CARD,\
    SCARD_UNPOWER_CARD as _SCARD_UNPOWER_CARD,\
    SCARD_EJECT_CARD as _SCARD_EJECT_CARD,\
    SCARD_S_SUCCESS as _SCARD_S_SUCCESS,\
    INFINITE as _INFINITE,\
    SCARD_STATE_ATRMATCH as _SCARD_STATE_ATRMATCH,\
    SCARD_STATE_UNAWARE as _SCARD_STATE_UNAWARE,\
    SCARD_STATE_IGNORE as _SCARD_STATE_IGNORE,\
    SCARD_STATE_UNAVAILABLE as _SCARD_STATE_UNAVAILABLE,\
    SCARD_STATE_EMPTY as _SCARD_STATE_EMPTY,\
    SCARD_STATE_PRESENT as _SCARD_STATE_PRESENT,\
    SCARD_STATE_EXCLUSIVE as _SCARD_STATE_EXCLUSIVE,\
    SCARD_STATE_INUSE as _SCARD_STATE_INUSE,\
    SCARD_STATE_MUTE as _SCARD_STATE_MUTE,\
    SCARD_STATE_CHANGED as _SCARD_STATE_CHANGED,\
    SCARD_STATE_UNKNOWN as _SCARD_STATE_UNKNOWN,\
    SCARD_ATTR_ATR_STRING as _SCARD_ATTR_ATR_STRING,\
    SCARD_ATTR_VENDOR_IFD_SERIAL_NO as _SCARD_ATTR_VENDOR_IFD_SERIAL_NO,\
    SCARD_ATTR_VENDOR_NAME as _SCARD_ATTR_VENDOR_NAME,\
    SCardEstablishContext as _SCardEstablishContext,\
    SCardListReaders as _SCardListReaders,\
    SCardConnect as _SCardConnect,\
    SCardReconnect as _SCardReconnect,\
    SCardStatus as _SCardStatus,\
    SCardTransmit as _SCardTransmit,\
    SCardDisconnect as _SCardDisconnect,\
    SCardReleaseContext as _SCardReleaseContext,\
    SCardGetErrorMessage as _SCardGetErrorMessage,\
    SCardGetStatusChange as _SCardGetStatusChange,\
    SCardGetAttrib as _SCardGetAttrib


# Local application imports
from common.intlist import to_hstr as _to_hstr
from common.hstr import \
    to_intlist as _to_intlist,\
    minimum as _min,\
    clean as _clean
from iso7816 import CommandApdu, ResponseApdu, CommandCase, GET_RESPONSE as _GET_RESPONSE

# Enum definitions


class Scope(IntEnum):
    User = _SCARD_SCOPE_USER
    System = _SCARD_SCOPE_SYSTEM


class ShareMode(IntEnum):
    Shared = _SCARD_SHARE_SHARED
    Exclusive = _SCARD_SHARE_EXCLUSIVE
    Direct = _SCARD_SHARE_DIRECT


class Protocol(IntEnum):
    T0 = _SCARD_PROTOCOL_T0
    T1 = _SCARD_PROTOCOL_T1
    T0orT1 = _SCARD_PROTOCOL_T0 | _SCARD_PROTOCOL_T1
    Raw = _SCARD_PROTOCOL_RAW


class Disposition(IntEnum):
    LeaveCard = _SCARD_LEAVE_CARD
    ResetCard = _SCARD_RESET_CARD
    UnpowerCard = _SCARD_UNPOWER_CARD
    EjectCard = _SCARD_EJECT_CARD


class State(IntEnum):
    Unknown = 0x0001
    Absent = 0x0002
    Present = 0x0004
    Swallowed = 0x0008
    Powered = 0x0010
    Negotiable = 0x0020
    Specific = 0x0040


class Attribute(IntEnum):
    AtrString = _SCARD_ATTR_ATR_STRING
    IfdSerialNo = _SCARD_ATTR_VENDOR_IFD_SERIAL_NO
    VendorName = _SCARD_ATTR_VENDOR_NAME


# Exception definition

class PcscError(Exception):
    pass


# Functions definitions

def establish_context(dw_scope: Scope):
    """establish_context():
    """
    hresult, hcontext = _SCardEstablishContext(dw_scope)
    if hresult != _SCARD_S_SUCCESS:
        err = F'Failed to establish PC/SC context ({_SCardGetErrorMessage(hresult)})'
        logging.error(err)
        raise PcscError(err)

    logging.info(F"PC/SC context established with scope: {dw_scope.name}")
    return hcontext


def list_readers(hcontext, readergroups=None):
    """list_readers():
    """
    if readergroups is None:
        readergroups = []

    hresult, readers = _SCardListReaders(hcontext, readergroups)
    if hresult != _SCARD_S_SUCCESS:
        err = F'Failed to list readers ({_SCardGetErrorMessage(hresult)})'
        logging.error(err)
        raise PcscError(err)

    l = len(readers)
    match l:
        case 0:
            logging.info("No reader found")
        case 1:
            logging.info(F"Reader found: {readers[0]}")
        case _:
            logging.info(F"{l} readers found: {', '.join(readers)}")

    return readers


def list_readers_serialno(hcontext, readers):
    """list_readers_serialno():
    """
    serial_nos = []
    for r in readers:
        hcard, _ = connect(hcontext, r, ShareMode.Direct, Protocol.T0orT1)
        serial_nos.append(get_attribute(hcard, Attribute.IfdSerialNo))
        disconnect(hcard, Disposition.UnpowerCard)

    logging.info(
        F"Serial nos listed: {', '.join([F'{r} ({s})' for r,s in zip(readers, serial_nos)])}")
    return serial_nos


def connect(hcontext, reader, dw_share_mode: ShareMode, dw_preferred_protocols: Protocol):
    """connect():
    """
    hresult, hcard, dw_active_protocol = _SCardConnect(
        hcontext, reader, dw_share_mode, dw_preferred_protocols)
    if hresult != _SCARD_S_SUCCESS:
        err = F'Unable to connect ({_SCardGetErrorMessage(hresult)})'
        logging.error(err)
        raise PcscError(err)

    protocol = Protocol(dw_active_protocol)

    logging.info(F"Connected with share mode: {dw_share_mode.name}")
    logging.info(F"Connected with active protocol: {protocol.name}")

    return hcard, protocol


def reconnect(hcard, dw_share_mode: ShareMode, dw_preferred_protocols: Protocol, dw_initialization: Disposition):
    """reconnect():
    """
    hresult, dw_active_protocol = _SCardReconnect(
        hcard, dw_share_mode, dw_preferred_protocols, dw_initialization)

    if hresult != _SCARD_S_SUCCESS:
        err = F'Unable to reconnect ({_SCardGetErrorMessage(hresult)})'
        logging.error(err)
        raise PcscError(err)

    active_protocol = Protocol(dw_active_protocol)

    logging.info(F"Reconnected with share mode: {dw_share_mode.name}")
    logging.info(F"Reconnected with disposition: {dw_initialization.name}")
    logging.info(F"Reconnected with active protocol: {active_protocol.name}")

    return active_protocol


def status(hcard):
    """status():
    """
    hresult, reader, dw_state, dw_protocol, atr = _SCardStatus(hcard)

    if hresult != _SCARD_S_SUCCESS:
        err = F'Unable to get current reader status ({_SCardGetErrorMessage(hresult)})'
        logging.error(err)
        raise PcscError(err)

    protocol = Protocol(dw_protocol)
    states = []
    for scard_state in State:
        if dw_state & scard_state:
            states.append(scard_state)

    logging.info(F"Card ATR: {_to_hstr(atr)}")
    logging.info(F"Card connected with active protocol: {protocol.name}")
    logging.info(F"Card in reader: {reader}")
    l = len(states)
    match l:
        case 0:
            logging.info("No card state found")
        case 1:
            logging.info(F"Card state: {states[0].name}")
        case _:
            logging.info(
                F"{l} card states found: {', '.join([state.name for state in states])}")

    return reader, states, protocol, _to_hstr(atr)


def transmit(hcard, protocol: Protocol, command_apdu: CommandApdu) -> ResponseApdu:
    """transmit()
    """
    pio_send_pci = {Protocol.T0: _SCARD_PCI_T0,
                    Protocol.T1: _SCARD_PCI_T1}[protocol]

    hresult, response = _SCardTransmit(
        hcard, pio_send_pci, command_apdu.list())
    if hresult != _SCARD_S_SUCCESS:
        err = F'Failed to transmit ({_SCardGetErrorMessage(hresult)})'
        logging.error(err)
        raise PcscError(err)

    response_apdu = ResponseApdu.from_list(response)
    logging.debug(F"Tx->{command_apdu.str()}")
    logging.debug(F"Rx<-{response_apdu.str()}")

    return response_apdu


def send_apdu(hcard, protocol: Protocol, command_apdu: CommandApdu) -> ResponseApdu:
    """send_apdu()
    """
    match protocol:
        case Protocol.T0:
            match command_apdu.case():
                case CommandCase.Case1:
                    return _send_apdu_T0_case_1(hcard, command_apdu)

                case CommandCase.Case2s:
                    return _send_apdu_T0_case_2s(hcard, command_apdu)

                case CommandCase.Case2e:
                    raise PcscError('Case 2e not yet implemented for T=0')

                case CommandCase.Case3s:
                    return _send_apdu_T0_case_3s(hcard, command_apdu)

                case CommandCase.Case3e:
                    raise PcscError('Case 3e not yet implemented for T=0')

                case CommandCase.Case4s:
                    return _send_apdu_T0_case_4s(hcard, command_apdu)

                case CommandCase.Case4e:
                    raise PcscError('Case 4e not yet implemented for T=0')

        case Protocol.T1:
            return transmit(hcard, protocol, command_apdu)

        case _:
            raise PcscError(
                F'Unsupported protocol: {protocol.name}')


def _send_apdu_T0_case_1(hcard, command_apdu: CommandApdu) -> ResponseApdu:
    """send_apdu_T0_case_1():
    """
    logging.debug('Command APDU Case 1')
    return transmit(hcard, Protocol.T0, command_apdu)


def _send_apdu_T0_case_2s(hcard, command_apdu: CommandApdu) -> ResponseApdu:
    """send_apdu_T0_case_2s():
    """
    logging.debug('Command APDU Case 2S')
    response_apdu = transmit(hcard, Protocol.T0, command_apdu)

    match response_apdu.SW1(), response_apdu.SW2():
        case ('90', '00'):
            # Case 2S.1—Process completed: Ne accepted
            return response_apdu

        case ('67', '00'):
            # Case 2S.2—Process aborted: Ne definitively not accepted
            return response_apdu

        case ('6C', _):
            # Case 2S.3—Process aborted; Ne not accepted, Na indicated
            return transmit(hcard, Protocol.T0, command_apdu.update_Le(response_apdu.SW2()))

        case (sw1, _) if sw1.startswith('9'):
            # Case 2S.4—SW12 = '9XYZ', except for '9000'
            return response_apdu

        case _:
            return response_apdu


def _send_apdu_T0_case_3s(hcard, command_apdu: CommandApdu) -> ResponseApdu:
    logging.debug('Command APDU Case 3s')
    return transmit(hcard, Protocol.T0, command_apdu)


def _send_apdu_T0_case_4s(hcard, command_apdu: CommandApdu) -> ResponseApdu:
    logging.debug('Command APDU Case 4S')
    """send_apdu_T0_case_4s():
    """
    response_apdu = transmit(hcard, Protocol.T0, command_apdu)

    match (response_apdu.SW1(), response_apdu.SW2()):
        case ('64', _) | ('65', _) | ('66', _):
            # Case 4S.1—Process aborted
            return response_apdu

        case ('67', '00') | ('68', _) | ('69', _) | ('6A', _) | ('6C', _):
            # Case 4S.1—Process aborted
            return response_apdu

        case ('6B', '00'):
            # Case 4S.1—Process aborted
            return response_apdu

        case ('6D', '00'):
            # Case 4S.1—Process aborted
            return response_apdu

        case ('6E', '00'):
            # Case 4S.1—Process aborted
            return response_apdu

        case ('6F', '00'):
            # Case 4S.1—Process aborted
            return response_apdu

        case ('90', '00'):
            # Case 4S.2—Process completed
            # Send GET RESPONSE with same Le
            return _send_apdu_T0_case_2s(hcard, _GET_RESPONSE(command_apdu.CLA(), command_apdu.Le()))

        case ('61', sw2):
            # Case 4S.3—Process completed with information added
            Nx = int(sw2, 16)
            if command_apdu.Le() == '00':
                Ne = 0x100
            else:
                Ne = int(command_apdu.Le(), 16)

            Ne = min(Ne, Nx)
            return _send_apdu_T0_case_2s(hcard, _GET_RESPONSE(command_apdu.CLA(), F"{Ne:02X}"))

        case ('62', _) | ('63', _):
            # Case 4S.4—SW12 = '62XY' or '63XY'
            return response_apdu

        case (sw1, _) if sw1.startswith('9'):
            # Case 4S.4—SW12 = '9XYZ', except for '9000'
            return response_apdu

        case _:
            return response_apdu


def get_status_change(hcontext, reader_states=None, timeout=None):
    """get_status_change()
    """
    if reader_states is None:
        readers = list_readers(hcontext)

        reader_states = [(reader, _SCARD_STATE_UNAWARE)
                         for reader in readers]
    old_reader_states = list(
        map(_convert_reader_state_to_scard, reader_states))

    if timeout is None:
        timeout = _INFINITE

    hresult, new_reader_states = _SCardGetStatusChange(
        hcontext, timeout, old_reader_states)
    if hresult != _SCARD_S_SUCCESS:
        err = F'Failed to get new reader states ({_SCardGetErrorMessage(hresult)})'
        logging.error(err)
        raise PcscError(err)

    new_reader_states = list(
        map(_convert_reader_state_from_scard, new_reader_states))

    logging.info(F"new reader states: {', '.join(new_reader_states.name)}")
    return new_reader_states


def get_attribute(hcard, attribute: Attribute):
    hresult, value = _SCardGetAttrib(hcard, attribute.value)
    if hresult != _SCARD_S_SUCCESS:
        err = F'Failed to get attribute {attribute} ({_SCardGetErrorMessage(hresult)})'
        logging.error(err)
        raise PcscError(err)
    if attribute == Attribute.AtrString:
        return _to_hstr(value)
    elif attribute in [Attribute.IfdSerialNo, Attribute.VendorName]:
        return ''.join([chr(i) for i in value])
    else:
        return


def disconnect(hcard, dw_disposition: Disposition):
    """disconnect():
    """
    hresult = _SCardDisconnect(hcard, dw_disposition)
    if hresult != _SCARD_S_SUCCESS:
        err = F'Failed to disconnect ({_SCardGetErrorMessage(hresult)})'
        logging.error(err)
        raise PcscError(err)

    logging.info("Disconnected")


def release_context(hcontext):
    """release_context():
    """
    hresult = _SCardReleaseContext(hcontext)
    if hresult != _SCARD_S_SUCCESS:
        err = F'Failed to release PC/SC context: {_SCardGetErrorMessage(hresult)}'
        logging.error(err)
        raise PcscError(err)

    logging.info("PC/SC context released")

#
# Helper functions
#


# def _compare_reader_states(old_reader_states, new_reader_states):
#     return [new_reader_state for new_reader_state in new_reader_states if new_reader_state[1] & _SCARD_STATE_CHANGED]


def _convert_reader_state_from_scard(reader_state):
    if len(reader_state) == 3:
        reader_state + (_to_hstr(reader_state[2]),)
    else:
        return reader_state


def _convert_reader_state_to_scard(reader_state):
    return reader_state[0:3]



"""commands.py
"""

# Standard library imports
from enum import Enum, IntEnum
from typing import Union

# Third party imports

# Local application imports
from common.hstr import clean as _clean
from globalplatform.encodings import SecureMessaging, CLA as GP_CLA
from iso7816 import CommandApdu, CommandField, CLA as ISO_CLA, Chaining


# Enum Definitions


class GetDataObject(Enum):
    ListOfApplications = '2F00'
    IssuerIdentificationNumber = '0042'
    CardImageNumber = '0045'
    CardData = '0066'
    KeyInformationTemplate = '00E0'
    CardCapabilityInformation = '0067'
    CurrentSecurityLevel = '00D3'
    SecurityDomainManagerURL = '5F50'
    ConfirmationCounter = '00C2'
    SequenceCounterOfTheDefaultKeyVersionNumber = '00C1'
    CardProductionLifeCycle = '9F7F'


class FileOccurrence(IntEnum):
    FirstOrOnlyOccurrence = 0x0
    NextOccurrence = 0x2


class ApplicationIdentifier(Enum):
    GlobalPlatformSecurityDomain = 'A000000151000000'
    Default = ''


# Command APDUs defined by GP


def GET_DATA(secure_messaging: SecureMessaging, logical_channel: int, tag: Union[str, GetDataObject]) -> CommandApdu:
    """GET_DATA: generate APDU for GET DATA command
    """
    apdu = {CommandField.Class: GP_CLA(secure_messaging, logical_channel).str(),
            CommandField.Instruction: 'CA',
            CommandField.Le: '00'}

    if tag in GetDataObject:
        apdu[CommandField.P1] = tag.value[0:2]
        apdu[CommandField.P2] = tag.value[2:4]
    else:
        apdu[CommandField.P1] = _clean(tag[0:2])
        apdu[CommandField.P2] = _clean(tag[2:4])

    return CommandApdu.from_dict(apdu)


def SELECT(logical_channel: int, file_occurrence: FileOccurrence, application_identifier: Union[str, ApplicationIdentifier]) -> CommandApdu:
    apdu_dict = {CommandField.Class: ISO_CLA(SecureMessaging.No, Chaining.LastOrOnly, logical_channel).str(),
                 CommandField.Instruction: 'A4',
                 CommandField.P1: '04',
                 CommandField.P2: F"{file_occurrence:02X}",
                 CommandField.Le: '00'}

    if application_identifier in ApplicationIdentifier:
        if application_identifier != ApplicationIdentifier.Default:
            apdu_dict[CommandField.Data] = application_identifier.value
    else:
        apdu_dict[CommandField.Data] = _clean(
            application_identifier, 'SELECT()', 'application_identifier')

    return CommandApdu.from_dict(apdu_dict)



"""commands.py
"""

# Standard library imports
from enum import Enum
from typing import Union

# Third party imports

# Local application imports
from common.ber import encode
from common.hstr import clean as _clean
from iso7816.apdu import CommandApdu, CommandField


# Enum Definitions


class ApplicationIdentifier(Enum):
    Mastercard = 'A0000000041010'
    Visa = 'A0000000031010'
    Maestro = 'A00000000410103060'
    EVC = 'A0000008864040'
    EVC_test = 'A0000008863030'
    Dummy = 'A0000008862020'
    Dummy_test = 'A0000008861010'


class GetDataObject(Enum):
    ATC = '9F36'
    LastOnlineATCRegister = '9F13'
    PINTryCounter = '9F17'
    LogFormat = '9F4f'


# Command APDUs defined by EMV


def SELECT(aid: Union[str, ApplicationIdentifier]) -> CommandApdu:
    """SELECT(): generate CommandApdu for SELECT command
    """
    apdu = {CommandField.Header: '00A40400', CommandField.Le: '00'}

    if aid in ApplicationIdentifier:
        apdu[CommandField.Data] = aid.value
    else:
        apdu[CommandField.Data] = _clean(aid, "emv.SELECT()", 'aid')

    return CommandApdu.from_dict(apdu)


def GET_PROCESSING_OPTIONS(pdol: str = '') -> CommandApdu:
    """GET_PROCESSING_OPTIONS(): generate APDU for GET PROCESSING OPTIONS command
    """
    apdu = {CommandField.Header: '80A80000',
            CommandField.Data: encode('83', pdol),
            CommandField.Le: '00'}

    return CommandApdu.from_dict(apdu)


def GPO(pdol: str) -> CommandApdu:
    """GPO(): generate APDU for GET PROCESSING OPTIONS command
    """
    return GET_PROCESSING_OPTIONS(pdol)


def READ_RECORD(SFI: int, record: int) -> CommandApdu:
    """READ_RECORD(): generate APDU for READ RECORD command
    """
    apdu = {CommandField.Class: '00',
            CommandField.Instruction: 'B2',
            CommandField.P1: F"{record:02X}",
            CommandField.P2: F"{SFI*8+4:02X}",
            CommandField.Le: '00'}

    return CommandApdu.from_dict(apdu)


def GET_DATA(tag: Union[str, GetDataObject]) -> CommandApdu:
    """GET_DATA: generate APDU for GET DATA command
    """
    apdu = {CommandField.Class: '80',
            CommandField.Instruction: 'CA',
            CommandField.Le: '00'}

    if tag in GetDataObject:
        apdu[CommandField.P1] = tag.value[0:2]
        apdu[CommandField.P2] = tag.value[2:4]
    else:
        apdu[CommandField.P1] = _clean(tag[0:2])
        apdu[CommandField.P2] = _clean(tag[2:4])

    return CommandApdu.from_dict(apdu)



"""encodings.py
"""

# Standard library imports
from enum import Enum

# Third party imports

# Local application imports
from common.hstr import clean as _clean


# Enum Definitions

class SecureMessaging(Enum):
    No = 'No secure messaging'
    GlobalPlatform = 'Secure messaging - GlobalPlatform proprietary'
    Iso7816 = 'Secure messaging - ISO7816 standard, command header not processed (no C-MAC)'
    Iso7816_CMAC = 'Secure messaging - ISO7816 standard, command header authenticated (C-MAC)'


class CardPersonalizationLifeCycleData(Enum):
    ICFabricator = "IC Fabricator"
    ICType = "IC Type"
    OperatingSystemIdentifier = "Operating System Identifier"
    OperatingSystemReleaseDate = "Operating System Release Date"
    OperatingSystemReleaseLevel = "Operating System Release Level"
    ICFabricationDate = "IC Fabrication Date"
    ICSerialNumber = "IC Serial Number"
    ICBatchIdentifier = "IC Batch Identifier"
    ICModuleFabricator = "IC Module Fabricator"
    ICModulePackagingDate = "IC Module Packaging Date"
    ICCManufacturer = "ICC Manufacturer"
    ICEmbeddingDate = "IC Embedding Date"
    ICPrePersonalizer = "IC Pre-Personalizer"
    ICPrePersonalizationDate = "IC Pre-Personalization Date"
    ICPrePersonalizationEquipmentIdentifier = "IC Pre-Personalization Equipment Identifier"
    ICPersonalizer = "IC Personalizer"
    ICPersonalizationDate = "IC Personalization Date"
    ICPersonalizationEquipmentIdentifier = "IC Personalization Equipment Identifier"

# Class Definitions


class CLA:
    def __init__(self, secure_messaging: SecureMessaging, logical_channel: int):
        if (logical_channel < 0) or (logical_channel > 19):
            raise ValueError(
                F'Logical channel number out of bound: should be in [0;19], received {logical_channel}')

        self.__logical_channel = logical_channel
        self.__secure_messaging = secure_messaging

    @classmethod
    def from_value(cls, CLA: int):
        if (CLA < 0x00) or (CLA > 0xFF):
            raise ValueError(
                F'Class byte should out of bound: should be 2 bytes, received {CLA:X}')

        secure_messaging: SecureMessaging
        logical_channel: int

        # Testing b7
        if CLA & 0x40 == 0x00:
            # CLA Byte Coding according to Table 11-11
            # Testing b4-b3
            match CLA & 0x0C:
                case 0x00:
                    secure_messaging = SecureMessaging.No
                case 0x04:
                    secure_messaging = SecureMessaging.GlobalPlatform
                case 0x08:
                    secure_messaging = SecureMessaging.Iso7816
                case 0x0C:
                    secure_messaging = SecureMessaging.Iso7816_CMAC
                case _:
                    # This should not happen
                    raise ArithmeticError(
                        F'CLA.from_value() error: CLA = {CLA:02X}')

            logical_channel = CLA & 0x03

        else:
            # CLA Byte Coding according to Table 11-12
            # Testing b6
            if CLA & 0x20 == 0x00:
                secure_messaging = SecureMessaging.No
            else:
                secure_messaging = SecureMessaging.GlobalPlatform

            logical_channel = CLA & 0x0F

        return cls(secure_messaging, logical_channel)

    @classmethod
    def from_string(cls, CLA: str):
        return cls.from_value(cls, int(_clean(CLA), 16))

    def value(self) -> int:
        CLA = 0x00

        if self.__logical_channel <= 3:
            # CLA Byte Coding according to Table 11-11
            CLA += 0x80 + self.__logical_channel
            match self.__secure_messaging:
                case SecureMessaging.No:
                    CLA += 0x00
                case SecureMessaging.GlobalPlatform:
                    CLA += 0x04
                case SecureMessaging.Iso7816:
                    CLA += 0x08
                case SecureMessaging.Iso7816_CMAC:
                    CLA += 0x0C

        else:
            # CLA Byte Coding according to Table 11-12
            CLA += 0xC0 + self.__logical_channel
            if self.__secure_messaging != SecureMessaging.No:
                CLA += 0x20

        return CLA

    def str(self) -> str:
        return F"{self.value():02X}"


class CPLC:
    def __init__(self, data: str):
        cplc_data = _clean(
            data, "globalplatform.encodings.CPLC.__init__()", "data")
        if len(cplc_data) != 84:
            raise ValueError(
                F"CPLC data should be 42 bytes, received {len(cplc_data)//2} bytes: {cplc_data}")
        else:
            self.__cplc_data = cplc_data

    def get_field(self, data: CardPersonalizationLifeCycleData):
        pos, length = _CPLC_DATA_DEFINITION[data]
        return self.__cplc_data[2*pos:2*(pos+length)]

    def str(self):
        return self.__cplc_data


# Helper functions
_CPLC_DATA_DEFINITION = {
    CardPersonalizationLifeCycleData.ICFabricator: (0, 2),
    CardPersonalizationLifeCycleData.ICType: (2, 2),
    CardPersonalizationLifeCycleData.OperatingSystemIdentifier: (4, 2),
    CardPersonalizationLifeCycleData.OperatingSystemReleaseDate: (6, 2),
    CardPersonalizationLifeCycleData.OperatingSystemReleaseLevel: (8, 2),
    CardPersonalizationLifeCycleData.ICFabricationDate: (10, 2),
    CardPersonalizationLifeCycleData.ICSerialNumber: (12, 4),
    CardPersonalizationLifeCycleData.ICBatchIdentifier: (16, 2),
    CardPersonalizationLifeCycleData.ICModuleFabricator: (18, 2),
    CardPersonalizationLifeCycleData.ICModulePackagingDate: (20, 2),
    CardPersonalizationLifeCycleData.ICCManufacturer: (22, 2),
    CardPersonalizationLifeCycleData.ICEmbeddingDate: (24, 2),
    CardPersonalizationLifeCycleData.ICPrePersonalizer: (26, 2),
    CardPersonalizationLifeCycleData.ICPrePersonalizationDate: (28, 2),
    CardPersonalizationLifeCycleData.ICPrePersonalizationEquipmentIdentifier: (30, 4),
    CardPersonalizationLifeCycleData.ICPersonalizer: (34, 2),
    CardPersonalizationLifeCycleData.ICPersonalizationDate: (36, 2),
    CardPersonalizationLifeCycleData.ICPersonalizationEquipmentIdentifier: (38, 4),
}



from globalplatform.encodings import CardPersonalizationLifeCycleData, SecureMessaging, CLA, CPLC
from globalplatform.commands import GetDataObject, FileOccurrence, ApplicationIdentifier, GET_DATA, SELECT



"""apdu.py
"""

# Standard library imports
from enum import Enum

# Third party imports

# Local application imports
from common.hstr import clean as _clean, to_intlist as _to_intlist
from common.intlist import to_hstr as _to_hstr

# Enum Definitions


class CommandCase(Enum):
    Case1 = 'Case 1'
    Case2s = 'Case 2s'
    Case2e = 'Case 2e'
    Case3s = 'Case 3s'
    Case3e = 'Case 3e'
    Case4s = 'Case 4s'
    Case4e = 'Case 4e'


class CommandField(Enum):
    Header = 'Header'
    Body = 'Body'
    Class = 'Class'
    Instruction = 'Instruction'
    P1 = 'P1'
    P2 = 'P2'
    Lc = 'Lc'
    Data = 'Data'
    Le = 'Le'


class ResponseField(Enum):
    Body = 'Body'
    Trailer = 'Trailer'
    Data = 'Data'
    SW1 = 'SW1'
    SW2 = 'SW2'
    SW12 = 'SW12'


class ResponseProcessingState(Enum):
    Normal = 'Normal procesing'
    Warning = 'Warning processing'
    ExecutionError = 'Execution error'
    CheckingError = 'Checking error'

# Command and Response APDU classes


class CommandApdu:
    def __init__(self, apdu_str: str):
        self.__list = _to_intlist(
            apdu_str, "CommandApdu.__init__()", "apdu_str")

        apdu_length = len(self.__list)

        if apdu_length < 4:
            raise ValueError(
                F'Wrong Command APDU length: expected 4 or more bytes but received {apdu_length}')

        elif apdu_length == 4:
            self.__case = CommandCase.Case1

        elif apdu_length == 5:
            self.__case = CommandCase.Case2s

        elif self.__list[4] != 0x00:
            if apdu_length == 5 + self.__list[4]:
                self.__case = CommandCase.Case3s
            elif apdu_length == 5 + self.__list[4] + 1:
                self.__case = CommandCase.Case4s
            else:
                raise ValueError(
                    F'Wrong Command APDU length: expected {5+self.__list[4]} bytes for Case 3s or {5+self.__list[4]+1} bytes for Case 4s but received {apdu_length} bytes instead')

        else:
            # Extended length cases
            if apdu_length == 7:
                self.__case = CommandCase.Case2e

            elif apdu_length == 7 + self.__list[5] * 0x0100 + self.__list[6]:
                self.__case = CommandCase.Case3e

            elif apdu_length == 7 + self.__list[5] * 0x0100 + self.__list[6] + 2:
                self.__case = CommandCase.Case4e

            else:
                raise ValueError(
                    F'Wrong Command APDU length: expected {7 + self.__list[5] * 0x0100 + self.__list[6]} bytes for Case 3e or {7 + self.__list[5] * 0x0100 + self.__list[6] + 2} bytes for Case 4e but received {apdu_length} bytes instead')

    @classmethod
    def from_dict(cls, apdu: dict):
        # If 'apdu' includes a header, it will be used directly. If not, the header will be constructed with CLA, INS, P1, and P2
        if CommandField.Header in apdu:
            apdu_str = _clean(
                apdu[CommandField.Header], 'CommandApdu.from_dict()', 'apdu[CommandField.Header]')
        else:
            apdu_str = F"{_clean(apdu[CommandField.Class], 'CommandApdu.from_dict()', 'apdu[CommandField.Class]')}{_clean(apdu[CommandField.Instruction], 'CommandApdu.from_dict()', 'apdu[CommandField.Instruction]')}{_clean(apdu[CommandField.P1], 'CommandApdu.from_dict()', 'apdu[CommandField.P1]')}{_clean(apdu[CommandField.P2], 'CommandApdu.from_dict()', 'apdu[CommandField.P2]')}"

        # If 'apdu' includes a body, it will be used directly. If not, the body will be constructed with the optional DATA and Le fields
        if CommandField.Body in apdu:
            apdu_str += _clean(apdu[CommandField.Body],
                               'CommandApdu.from_dict()', 'apdu[CommandField.Body]')
        else:

            if CommandField.Data in apdu:
                _DATA = _clean(
                    apdu[CommandField.Data], 'CommandApdu.from_dict()', 'apdu[CommandField.Data]')
                apdu_str += F"{''.join([F'{l:02X}' for l in _Lc(_to_intlist(_DATA))])}{_DATA}"

            if CommandField.Le in apdu:
                apdu_str += _clean(apdu[CommandField.Le],
                                   "CommandApdu.__init__()", "apdu[CommandField.Le]")

        return cls(apdu_str)

    def list(self):
        return self.__list

    def str(self):
        return _to_hstr(self.__list)

    def get_field(self, field: CommandField):
        match field:
            case CommandField.Header:
                return _to_hstr(self.__list[0:4])

            case CommandField.Body:
                return _to_hstr(self.__list[4:])

            case CommandField.Class:
                return F"{self.__list[0]:02X}"

            case CommandField.Instruction:
                return F"{self.__list[1]:02X}"

            case CommandField.P1:
                return F"{self.__list[2]:02X}"

            case CommandField.P2:
                return F"{self.__list[3]:02X}"

            case CommandField.Lc:
                match self.__case:
                    case CommandCase.Case1 | CommandCase.Case2s | CommandCase.Case2e:
                        raise ValueError(
                            F'Wrong Command APDU field: {self.__case} has no Lc field')
                    case CommandCase.Case3s | CommandCase.Case4s:
                        return F"{self.__list[4]:02X}"
                    case CommandCase.Case3e | CommandCase.Case4e:
                        return _to_hstr(self.__list[4:7])

            case CommandField.Data:
                match self.__case:
                    case CommandCase.Case1 | CommandCase.Case2s | CommandCase.Case2e:
                        raise ValueError(
                            F'Wrong Command APDU field: {self.__case} has no Data field')
                    case CommandCase.Case3s:
                        return _to_hstr(self.__list[5:])
                    case CommandCase.Case4s:
                        return _to_hstr(self.__list[5:-1])
                    case CommandCase.Case3e:
                        return _to_hstr(self.__list[7:])
                    case CommandCase.Case4e:
                        return _to_hstr(self.__list[7:-2])

            case CommandField.Le:
                match self.__case:
                    case CommandCase.Case1 | CommandCase.Case3s | CommandCase.Case3e:
                        raise ValueError(
                            F'Wrong Command APDU field: {self.__case} has no Le field')
                    case CommandCase.Case2s | CommandCase.Case4s:
                        return F"{self.__list[-1]:02X}"
                    case CommandCase.Case2e | CommandCase.Case4e:
                        return _to_hstr(self.__list[-2:])

    def CLA(self):
        return self.get_field(CommandField.Class)

    def INS(self):
        return self.get_field(CommandField.Instruction)

    def P1(self):
        return self.get_field(CommandField.P1)

    def P2(self):
        return self.get_field(CommandField.P2)

    def Le(self):
        return self.get_field(CommandField.Le)

    def case(self):
        return self.__case

    def __str__(self):
        str = F"Command APDU ({self.__case.value}): CLA={self.CLA()}, INS={self.INS()}, P1={self.P1()}, P2={self.P2()}"

        match self.__case:
            case CommandCase.Case1:
                return str
            case CommandCase.Case2s | CommandCase.Case2e:
                return str + F", Le={self.Le()}"
            case CommandCase.Case3s | CommandCase.Case3e:
                return str + F", Lc={self.Lc()}, DATA={self.DATA()}"
            case CommandCase.Case4s | CommandCase.Case4e:
                return str + F", Lc={self.Lc()}, DATA={self.DATA()}, Le={self.Le()}"

    def update_Le(self, Le: str):
        match self.__case:
            case CommandCase.Case1 | CommandCase.Case3s | CommandCase.Case3e:
                raise ValueError(
                    F'Wrong Command APDU field: {self.__case} has no Le field')

            case CommandCase.Case2s | CommandCase.Case2e:
                return CommandApdu.from_dict({CommandField.Header: self.get_field(CommandField.Header),
                                              CommandField.Le: Le})

            case CommandCase.Case4s | CommandCase.Case4e:
                return CommandApdu.from_dict({CommandField.Header: self.get_field(CommandField.Header),
                                              CommandField.Data: self.get_field(CommandField.Data),
                                              CommandField.Le: Le})


class StatusBytes:
    def __init__(self, sw12: str):
        self.__list = _to_intlist(sw12, "StatusBytes.__init__()", "sw12")
        if len(self.__list) != 2:
            raise ValueError(F"Expected 2 bytes, received: {sw12}")

        first_digit = self.__list[0] >> 4
        if first_digit not in [0x6, 0x9]:
            raise ValueError(F"Expected 6XXX or 9XXX value, received: {sw12}")

        if self.__list[0] == 0x60:
            raise ValueError(F"60XX in valid, received: {sw12}")

        match self.__list:
            case [0x90, 0x00]:
                self.__state = ResponseProcessingState.Normal
                self.__meaning = 'No further qualification'
            case [0x61, _]:
                self.__state = ResponseProcessingState.Normal
                self.__meaning = 'SW2 encodes the number of data bytes still available'
            case [0x62, sw2]:
                self.__state = ResponseProcessingState.Warning
                self.__meaning = {0x00: 'No information given',
                                  0x81: 'Part of returned data may be corrupted',
                                  0x82: 'End of file or record reached before reading Ne bytes',
                                  0x83: 'Selected file deactivated',
                                  0x84: 'File control information not formatted according to ISO7816-4 5.3.3',
                                  0x85: 'Selected file in termination state',
                                  0x86: 'No input data available from a sensor on the card'}.get(sw2, 'State of non-volatile memory is unchanged')
            case [0x63, sw2]:
                self.__state = ResponseProcessingState.Warning
                self.__meaning = {0x00: 'No information given',
                                  0x81: 'File filled up by the last '}.get(sw2, 'State of non-volatile memory has changed(further qualification in SW2)')
            case [0x64, sw2]:
                self.__state = ResponseProcessingState.ExecutionError
                self.__meaning = {0x00: 'Execution error',
                                  0x01: 'Immediate response required by the card'}.get(sw2, 'State of non-volatile memory is unchanged (further qualification in SW2)')
            case [0x65, sw2]:
                self.__state = ResponseProcessingState.ExecutionError
                self.__meaning = {0x00: 'No information given',
                                  0x81: 'Memory failure'}.get(sw2, 'State of non-volatile memory has changed (further qualification in SW2)')
            case [0x66, _]:
                self.__state = ResponseProcessingState.ExecutionError
                self.__meaning = 'Security-related issues'
            case [0x67, 0x00]:
                self.__state = ResponseProcessingState.CheckingError
                self.__meaning = 'Wrong length; no further indication'
            case [0x68, sw2]:
                self.__state = ResponseProcessingState.CheckingError
                self.__meaning = {0x00: 'No information given',
                                  0x81: 'Logical channel not supported',
                                  0x82: 'Secure messaging not supported',
                                  0x83: 'Last command of the chain expected',
                                  0x84: 'Command chaining not supported'}.get(sw2, 'Functions in CLA not supported (further qualification in SW2)')
            case [0x69, sw2]:
                self.__state = ResponseProcessingState.CheckingError
                self.__meaning = {0x00: 'No information given',
                                  0x81: 'Command incompatible with file structure',
                                  0x82: 'Security status not satisfied',
                                  0x83: 'Authentication method blocked',
                                  0x84: 'Reference data not usable',
                                  0x85: 'Conditions of use not satisfied',
                                  0x86: 'Command not allowed (no current EF)',
                                  0x87: 'Expected secure messaging data objects missing',
                                  0x88: 'Incorrect secure messaging data objects'}.get(sw2, 'Command not allowed (further qualification in SW2)')
            case [0x6A, _]:
                self.__state = ResponseProcessingState.CheckingError
                self.__meaning = 'Wrong parameters P1-P2 (further qualification in SW2)'
            case [0x6B, 0x00]:
                self.__state = ResponseProcessingState.CheckingError
                self.__meaning = 'Wrong parameters P1-P2'
            case [0x6C, _]:
                self.__state = ResponseProcessingState.CheckingError
                self.__meaning = 'Wrong Le field; SW2 encodes the exact number of available data bytes'
            case [0x6D, 0x00]:
                self.__state = ResponseProcessingState.CheckingError
                self.__meaning = 'Instruction code not supported or invalid'
            case [0x6E, 0x00]:
                self.__state = ResponseProcessingState.CheckingError
                self.__meaning = 'Class not supported'
            case [0x6F, 0x00]:
                self.__state = ResponseProcessingState.CheckingError
                self.__meaning = 'No precise diagnosis'

    def state(self):
        return self.__state

    def meaning(self):
        return self.__meaning


class ResponseApdu:
    def __init__(self, apdu_str: str):
        self.__list = _to_intlist(
            apdu_str, "ResponseApdu.__init__()", "apdu_str")

    @classmethod
    def from_list(cls, response: list[int]):
        return cls(_to_hstr(response))

    def list(self):
        return self.__list

    def str(self):
        return _to_hstr(self.__list)

    def get_field(self, field: ResponseField):
        match field:
            case ResponseField.Trailer | ResponseField.SW12:
                return _to_hstr(self.__list[-2:])

            case ResponseField.SW1:
                return F"{self.__list[-2]:02X}"

            case ResponseField.SW2:
                return F"{self.__list[-1]:02X}"

            case ResponseField.Body | ResponseField.Data:
                if len(self.__list) > 2:
                    return _to_hstr(self.__list[0:-2])
                else:
                    raise ValueError(
                        'Wrong Response APDU field: no response body')

    def data(self):
        return self.get_field(ResponseField.Data)

    def SW12(self):
        return self.get_field(ResponseField.SW12)

    def SW1(self):
        return self.get_field(ResponseField.SW1)

    def SW2(self):
        return self.get_field(ResponseField.SW2)

    def StatusBytes(self):
        return StatusBytes(self.get_field(ResponseField.SW12))

    def __str__(self):
        return _to_hstr(self.__list)

#
# Helper functions
#


def _Lc(data: list[int]):
    Lc = len(data)
    if Lc < 256:
        return [Lc]

    if Lc < 65536:
        return [0x00, Lc >> 8, Lc & 0x00FF]

    raise ValueError(F'Data length should be <65,536: actual length = {Lc}')



"""commands.py
"""

# Standard library imports
from enum import IntEnum

# Third party imports

# Local application imports
from common.hstr import clean as _clean
from iso7816.apdu import CommandField, CommandApdu
from iso7816.encodings import SecureMessaging, Chaining, CLA

# Enum Definitions


class Selection(IntEnum):
    SelectMForDForEF = 0x00
    SelectChildDF = 0x01
    SelectEFUnderCurrentDF = 0x02
    SelectParentDFOfCurrentDF = 0x03
    SelectByDFName = 0x04
    SelectFromMF = 0x08
    SelectFromCurrentDF = 0x09


class FileOccurrence(IntEnum):
    FirstOrOnlyOccurrence = 0x0
    LastOccurrence = 0x1
    NextOccurrence = 0x2
    PreviousOccurrence = 0x3


class FileControlInformation(IntEnum):
    ReturnFCITemplate = 0x00
    ReturnFCPTempalte = 0x10
    ReturnFMDTemplate = 0x20
    NoResponseOrProprietary = 0x30
#
# Commands for interchange
#


def GET_RESPONSE(CLA: str, Le: str):
    return CommandApdu.from_dict({CommandField.Class: CLA, CommandField.Instruction: 'C0', CommandField.P1: '00', CommandField.P2: '00', CommandField.Le: Le})


def SELECT(secure_messaging: SecureMessaging, chaining: Chaining, logical_channel: int, selection: Selection, file_occurrence: FileOccurrence, fci: FileControlInformation, data: str = None, Le: str = None) -> CommandApdu:
    apdu_dict = {CommandField.Class: CLA(secure_messaging, chaining, logical_channel).str(),
                 CommandField.Instruction: 'A4'}

    apdu_dict[CommandField.P1] = F"{selection:02x}"
    apdu_dict[CommandField.P2] = F"{file_occurrence + fci:02X}"

    if data is not None:
        apdu_dict[CommandField.Data] = _clean(data, 'SELECT()', 'data')

    if Le is not None:
        apdu_dict[CommandField.Le] = _clean(Le, 'SELECT()', 'Le')

    return CommandApdu.from_dict(apdu_dict)
